"""
This module contains the abstract class to implement process metrics.
"""

from pydriller import RepositoryMining

class ProcessMetric:
    """
    Abstract class to implement process metrics
    """

    def __init__(self, path_to_repo: str,
                 from_commit: str,
                 to_commit: str):
        """
        :path_to_repo: path to a single repo
        :to_commit: the SHA of the commit to stop counting. If None, the
            analysis starts from the latest commit
        :from_commit: the SHA of the commit to start counting. If None, the
            analysis ends to the first commit
        """
        if not from_commit or not to_commit:
            raise TypeError

        self.repo_miner = RepositoryMining(path_to_repo, single=from_commit)

        if from_commit != to_commit:
            self.repo_miner = RepositoryMining(path_to_repo=path_to_repo,
                                               from_commit=from_commit,
                                               to_commit=to_commit,
                                               reversed_order=True)

    def count(self):
        """
        Implement the main functionality of the metric
        """
        return 0



# Copyright 2018 Davide Spadini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pydriller.repository_mining import RepositoryMining
import logging

logging.basicConfig(format='%(asctime)s - %(levelname)s - %(message)s', level=logging.INFO)


def test_should_visit_ascendent_order():
    lc = list(RepositoryMining('test-repos/small_repo',
                               reversed_order=False).traverse_commits())
    assert len(lc) == 5
    assert lc[0].hash == 'a88c84ddf42066611e76e6cb690144e5357d132c'
    assert lc[1].hash == '6411e3096dd2070438a17b225f44475136e54e3a'
    assert lc[2].hash == '09f6182cef737db02a085e1d018963c7a29bde5a'
    assert lc[3].hash == '1f99848edadfffa903b8ba1286a935f1b92b2845'
    assert lc[4].hash == 'da39b1326dbc2edfe518b90672734a08f3c13458'


def test_should_visit_descendent_order():
    lc = list(RepositoryMining('test-repos/small_repo',
                               reversed_order=True).traverse_commits())
    assert len(lc) == 5
    assert lc[0].hash == 'da39b1326dbc2edfe518b90672734a08f3c13458'
    assert lc[1].hash == '1f99848edadfffa903b8ba1286a935f1b92b2845'
    assert lc[2].hash == '09f6182cef737db02a085e1d018963c7a29bde5a'
    assert lc[3].hash == '6411e3096dd2070438a17b225f44475136e54e3a'
    assert lc[4].hash == 'a88c84ddf42066611e76e6cb690144e5357d132c'


def test_should_visit_descendent_order_with_filters():
    lc = list(RepositoryMining('test-repos/small_repo',
                               from_commit='1f99848edadfffa903b8ba1286a935f1b92b2845',
                               to_commit='6411e3096dd2070438a17b225f44475136e54e3a',
                               reversed_order=True).traverse_commits())
    assert len(lc) == 3
    assert lc[0].hash == '1f99848edadfffa903b8ba1286a935f1b92b2845'
    assert lc[1].hash == '09f6182cef737db02a085e1d018963c7a29bde5a'
    assert lc[2].hash == '6411e3096dd2070438a17b225f44475136e54e3a'


def test_should_visit_descendent_order_with_filters_reversed():
    lc = list(RepositoryMining('test-repos/small_repo',
                               from_commit='6411e3096dd2070438a17b225f44475136e54e3a',
                               to_commit='1f99848edadfffa903b8ba1286a935f1b92b2845',
                               reversed_order=True).traverse_commits())
    assert len(lc) == 3
    assert lc[0].hash == '1f99848edadfffa903b8ba1286a935f1b92b2845'
    assert lc[1].hash == '09f6182cef737db02a085e1d018963c7a29bde5a'
    assert lc[2].hash == '6411e3096dd2070438a17b225f44475136e54e3a'


"""
Module that calculates the number of normalized added and deleted lines of a
file.
"""
from pydriller import ModificationType
from pydriller.metrics.process.process_metric import ProcessMetric


class LinesCount(ProcessMetric):
    """
    This class is responsible to implement the following metrics:
    * Added Lines: the number of added lines in the evolution period
        [from_commit, to_commit]
    * Deleted Lines: the number of deleted lines in the evolution period
        [from_commit, to_commit]
    """

    def count(self):
        """
        Calculate the number of normalized (by the total number of added and
        deleted lines) added and deleted lines per each modified file in
        'to_commit', returning a dictionary:
        {
          filepath: {
            added: int,
            removed: int
          }
        }

        :return: dict of total added and deleted lines per modified file
        """
        renamed_files = {}
        files = {}

        for commit in self.repo_miner.traverse_commits():

            for modified_file in commit.modifications:

                filepath = renamed_files.get(modified_file.new_path,
                                             modified_file.new_path)

                if modified_file.change_type == ModificationType.RENAME:
                    renamed_files[modified_file.old_path] = filepath

                if filepath not in files:
                    files[filepath] = {'added': 0, 'removed': 0}
                files[filepath]['added'] += modified_file.added
                files[filepath]['removed'] += modified_file.removed

        return files



"""
Module that calculates the number of developers that contributed to each
modified file in the repo in a given time range.

See https://dl.acm.org/doi/10.1145/2025113.2025119
"""
from pydriller import ModificationType
from pydriller.metrics.process.process_metric import ProcessMetric


class ContributorsCount(ProcessMetric):
    """
    This class is responsible to implement the following metrics:
    * Contributors Count: measures the number of contributors who modified a
      file.
    * Minor Contributors Count: measures the number of contributors who
      authored less than 5% of code of a file.
    """

    def count(self):
        """
        Return the number of contributors who modified a file and those that
        authored less than 5% of code for each modified file in the repository
        in the provided time range [from_commit, to_commit]

        :return: dict
            {filepath: {
                   contributors_count: int,
                   minor_contributors_count: int
                 }
            }
        of number of contributors for each modified file
        """
        renamed_files = {}
        files = {}

        for commit in self.repo_miner.traverse_commits():

            for modified_file in commit.modifications:

                filepath = renamed_files.get(modified_file.new_path,
                                             modified_file.new_path)

                if modified_file.change_type == ModificationType.RENAME:
                    renamed_files[modified_file.old_path] = filepath

                author = commit.author.email.strip()
                lines_authored = modified_file.added + modified_file.removed

                files[filepath] = files.get(filepath, {})
                files[filepath][author] = files[filepath].get(
                    author, 0) + lines_authored

        for path, contributions in list(files.items()):
            total = sum(contributions.values())
            if total == 0:
                del files[path]
            else:
                contributors_count = len(contributions.values())
                minor_contributors_count = sum(1
                                               for v in contributions.values()
                                               if v/total < .05)

                files[path] = {
                    'contributors_count': contributors_count,
                    'minor_contributors_count': minor_contributors_count
                }

        return files



from pathlib import Path

import pytest

from pydriller.metrics.process.lines_count import LinesCount

TEST_DATA = [
   ('test-repos/pydriller', '.gitignore', 'ab36bf45859a210b0eae14e17683f31d19eea041', 'ab36bf45859a210b0eae14e17683f31d19eea041',
      {'added': 197, 'removed': 0}),

   ('test-repos/pydriller', 'domain/modification.py', 'fdf671856b260aca058e6595a96a7a0fba05454b', 'ab36bf45859a210b0eae14e17683f31d19eea041',
      {'added': 49, 'removed': 1})
]

@pytest.mark.parametrize('path_to_repo, filepath, from_commit, to_commit, expected', TEST_DATA)
def test(path_to_repo, filepath, from_commit, to_commit, expected):
    metric = LinesCount(path_to_repo=path_to_repo,
                        from_commit=from_commit,
                        to_commit=to_commit)

    count = metric.count()
    filepath = str(Path(filepath))
    assert count[filepath] == expected
